#include "mcmc_density_estimation.h"

MCMC_Density_Estimation::MCMC_Density_Estimation( int argc, char **argv )
      : Basis_Fn_Regression( argc, argv )
{
   alpha = 1;
   log_alpha = 0;
   beta_distn_alpha = 1;
   beta_distn_beta = 1;
   random_basis_fns = 1;
   *Best_Changepoints_Filename = 0;
   *Input_Changepoints_Filename = 0;
   for( int i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
	 cout << "randombasisfns=" << random_basis_fns << endl;
	 cout << "dirichletparameter=" << alpha << endl;
	 cout << "betaparameter1=" << beta_distn_alpha << endl;
	 cout << "betaparameter2=" << beta_distn_beta << endl;
         cout << "cpinputfile=" << Input_Changepoints_Filename << "   (File which initial partition is read in from. Each row is a changepoint)" << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "randombasisfns=" ) )
         StringToBool( &random_basis_fns, argv[i] );
      if( InitialStringMatch( argv[i], "dirichletparameter=" ) )
	 StringToDouble( &alpha, argv[i] );
      if( InitialStringMatch( argv[i], "betaparameter1=" ) )
	 StringToDouble( &beta_distn_alpha, argv[i] );
      if( InitialStringMatch( argv[i], "betaparameter2=" ) )
	 StringToDouble( &beta_distn_beta, argv[i] );
      if( InitialStringMatch( argv[i], "cpinputfile=" ) )
         StringToString( Input_Changepoints_Filename, argv[i] );
   }
   initialise();
   if( print_to_screen )
   {
      cout << "Density estimation with Dirichlet parameter = " << alpha << endl;
      cout << "Base measure is Beta(" << beta_distn_alpha << "," << beta_distn_beta << ")" << endl;
   }

}

MCMC_Density_Estimation::MCMC_Density_Estimation( gsl_matrix* X1, gsl_matrix* Y1 )
        : Basis_Fn_Regression( X1, Y1 )
{
   initialise();
}

MCMC_Density_Estimation::MCMC_Density_Estimation( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
        : Basis_Fn_Regression( X1, Y1, D1 )
{
   initialise();
}

MCMC_Density_Estimation::~MCMC_Density_Estimation()
{
   zap_vector( num_cps_probs );
}

Regression* MCMC_Density_Estimation::create_new_regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
{
   Regression* Rptr = new MCMC_Density_Estimation( X1, Y1, D1 );
   return Rptr;
}

void MCMC_Density_Estimation::initialise()
{
   if( !strlen( PredictionFilename ) )
   {
      char* new_file = PredictionFilename;
      strcpy( new_file, TargetFilename );
      string insert = "_preds_";
      insert_in_string_before_dot( new_file, &insert );
   }
   if( !strlen( Best_Changepoints_Filename ) )
   {
      char* new_file = Best_Changepoints_Filename;
      strcpy( new_file, TargetFilename );
      string insert = "_map_changepoints_";
      insert_in_string_before_dot( new_file, &insert );
   }
   ADD_BASIS = DEL_BASIS = MOVE_BASIS = -INT_MAX;
   log_alpha = log( alpha );
   ml_const = 0;
   double total_volume = 1, transformer = 1;
   if( normalise_predictors != 1 )
      for( int j = 0; j < p; j++ )
	 total_volume *= prior_measure( x_min[j], x_max[j] );
   if( normalise_predictors )
   {
      for( int j = 0; j < p; j++ )
         transformer *= x_scale[ j ];
   }
   for( int i = 0; i < regression_n; i++ )
   {
       ml_const -= log( alpha * total_volume + i );
       if( logistic_transform_predictors )
	  ml_const += log(1-gsl_matrix_get(X,i,0)) + log(gsl_matrix_get(X,i,0));
   }
   ml_const -= regression_n * log( transformer );
   ml_const_store = ml_const;
   dirichlet_normaliser = (alpha * total_volume + regression_n );
   storing_basis_function_coefficients = false;
   if( random_basis_fns )
   {
      ADD_BASIS = add_move( "Add change point." );
      DEL_BASIS = add_move( "Delete change point." );
      trans_dimensional_move_types.push_back( ADD_BASIS );
      trans_dimensional_move_types.push_back( DEL_BASIS );
   }
      MOVE_BASIS = add_move( "Move change point." );
      integer_parameter_pointers.push_back( Basis_Fn::get_Basis_Fn_count_pointer() );
      //   }

   if( strlen( Input_Changepoints_Filename ) )
   {
       gsl_matrix* input_changepoints = scan_data( Input_Changepoints_Filename );
       if( input_changepoints )
       {
          for( unsigned int ii = 0; ii < input_changepoints->size1; ii++ )
             if( input_changepoints->size2 < 4 || !gsl_matrix_get( input_changepoints, ii, 3 ) )//not unremovable
             {
                double cp = gsl_matrix_get( input_changepoints, ii, 1 );
                int comp = gsl_matrix_get( input_changepoints, ii, 0 );
                if( normalise_predictors == 1 ) cp -= x_min[ comp ];                
                if( normalise_predictors ) cp /= x_scale[ comp ];
                add_basis_function( cp, comp, 0, 1 );
             }
          gsl_matrix_free( input_changepoints );
       }
   }
       
   number_removable_basis_fns = get_number_of_basis_fns() - number_unremovable_basis_functions;
   max_num_cps = max_num_splines > 0 ? max_num_splines + number_unremovable_basis_functions : 10000;
   num_cps_probs_length = max_num_cps + 1;
   num_cps_probs = new double[ num_cps_probs_length ];
   for( int j = 0; j < num_cps_probs_length; j++ )
      num_cps_probs[ j ] = 0;
   calculate_log_posterior();
   max_log_likelihood = log_likelihood = proposed_log_likelihood;
   partition_point_count = proposed_partition_point_count;
   partition_volumes = proposed_partition_volumes;
   partition_measures = proposed_partition_measures;
   partition_starting_points = proposed_partition_starting_points;
   partition_mls = proposed_partition_mls;
   partition_posterior_means = proposed_partition_posterior_means;
   set_basis_coeffs();
   update_posterior_maximizing_model();
//   if( !storing_basis_function_coefficients ) //
   make_predictions();
   write_predictions();
}

void MCMC_Density_Estimation::move()
{
   int num_possible_deletions = get_number_of_basis_fn_deletions_possible();
   int num_bases = get_number_of_basis_fns();
   if( move_type == ADD_BASIS )
   {
      if( num_bases == max_num_cps || !room_for_new_basis_fn() )
      {
         move_type = DEL_BASIS;
         if( max_num_cps == total_num_knot_point_candidates + 1 )
            LogProposalRatio -= LOG_THREE;
         else
            LogProposalRatio -= LOG_TWO;
      }
      else if( num_bases == max_num_cps - 1 )
         if( max_num_cps == total_num_knot_point_candidates + 1 )
            LogProposalRatio += LOG_THREE;
         else
            LogProposalRatio += LOG_TWO;
      else if( num_possible_deletions == 0 )
         LogProposalRatio -= LOG_THREE;
   }
   else if( move_type == DEL_BASIS )
   {
      if( num_possible_deletions == 0 )
      {
         move_type = ADD_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
      else if( num_possible_deletions == 1 )
         LogProposalRatio += LOG_THREE;
      else if( num_bases == max_num_cps )
      {
         if( max_num_cps == total_num_knot_point_candidates + 1 )
            LogProposalRatio -= LOG_THREE;
         else
            LogProposalRatio -= LOG_TWO;
      }
   }
   else if( move_type == MOVE_BASIS )
   {
      if( num_possible_deletions == 0 )
      {
         move_type = ADD_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
      else if( num_bases == total_num_knot_point_candidates + 1 )
      {
         move_type = DEL_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
   }
   propose_move();
}

void MCMC_Density_Estimation::propose_move()
{
   Basis_to_the_left = Basis_to_the_right = Removed_Basis = NULL;
   if( move_type == ADD_BASIS )
   {
      add_random_basis_function();
   }
   else if( move_type == DEL_BASIS )
   {
      remove_random_basis_function();
   }
   else if( move_type == MOVE_BASIS )
   {
      remove_random_basis_function();
      add_random_basis_function();
   }
   //   calculate_log_likelihood();
   //   proposed_log_posterior = proposed_log_likelihood;
   proposed_log_posterior = log_posterior;
   revise_log_posterior();
//   calculate_log_posterior();
//   switch_values( log_posterior, proposed_log_posterior );
}

void MCMC_Density_Estimation::execute_move()
{
   ml_const_store = ml_const;
   partition_point_count = proposed_partition_point_count;
   partition_volumes = proposed_partition_volumes;
   partition_measures = proposed_partition_measures;
   partition_starting_points = proposed_partition_starting_points;
   partition_posterior_means = proposed_partition_posterior_means;
   partition_mls = proposed_partition_mls;
   set_basis_coeffs();
}

void MCMC_Density_Estimation::undo_move()
{
   if( move_type == ADD_BASIS )
   {
      remove_basis_function( get_number_of_basis_fns() - 1 );
   }
   else if( move_type == DEL_BASIS )
   {
      add_basis_function( Removed_Basis, &removed_basis_position );
   }
   else if( move_type == MOVE_BASIS )
   {
      int col = get_number_of_basis_fns();
      if( removed_basis_position >= col )
         col--;
      add_basis_function( Removed_Basis, &removed_basis_position );
      remove_basis_function( get_number_of_basis_fns() - 1 );//col );
   }
   proposed_partition_point_count = partition_point_count;
   proposed_partition_volumes = partition_volumes;
   proposed_partition_measures = partition_measures;
   proposed_partition_starting_points = partition_starting_points;
   proposed_partition_posterior_means = partition_posterior_means;
   proposed_partition_mls = partition_mls;
   ml_const = ml_const_store;
}

void MCMC_Density_Estimation::post_move_updates()
{
//   if( !storing_basis_function_coefficients )//
//   set_basis_coeffs();
//      draw_betas_from_posterior();//automatically sets basis coefficients.
}

void MCMC_Density_Estimation::initialise_ergodic_averages()
{
   if( !ave_predictions_mx && predictions_mx_rows )
   {
      ave_predictions_mx_cols = predictions_update_cols + 1;
      if( variable_to_predict_over != second_variable_to_predict_over )
         ave_predictions_mx_cols++;
      ave_predictions_mx = new double*[ predictions_mx_rows ];
      for( int i = 0; i < predictions_mx_rows; i++ )
      {
         ave_predictions_mx[ i ] = new double[ ave_predictions_mx_cols ];
          for( int j = 0; j < ave_predictions_mx_cols; j++ )
            ave_predictions_mx[ i ][ j ] = 0;
      }
   }
//   if( !storing_basis_function_coefficients )//
   ml_const_store = ml_const;
}

void MCMC_Density_Estimation::update_ergodic_averages()
{
   num_cps_probs[ get_number_of_basis_fns() - number_unremovable_basis_functions ]++;
   update_knot_point_frequencies();
}

void MCMC_Density_Estimation::ModelPrintOut()
{
//   if( past_burn_in() )
//      double normaliser = get_current_ergodic_normaliser();
   cout << "Number of change_points = " << get_number_of_basis_fns() - number_unremovable_basis_functions << endl;
   draw_histogram( num_cps_probs, num_cps_probs_length );
}

void MCMC_Density_Estimation::make_predictions()
{
  if( mcmc_sampling && past_burn_in() && !( get_accept() || first_iteration_since_burn_in() ) )
      return;
   int last_col = 0;
   int i, j;//, shift = predict_on_data ? 0 : regression_n;
   bool extra_deleted_point = predict_on_data && deleted_x_point;
   bool predict_on_knot_point_candidates = !( n > regression_n || predict_on_data );
   predict_on_knot_point_candidates = false;
   if( predict_on_knot_point_candidates && !knot_point_candidates )
      return;
   double x_val = 0;
   if( !predictions_mx && ( n > regression_n || predict_on_data || predict_on_knot_point_candidates ) )
   {
      if( p == 1 )
         second_variable_to_predict_over = variable_to_predict_over;
      if( log_inputs || storing_basis_function_coefficients )
      {
         prediction_fn_value_col = 1;
         predictions_update_cols = 1;
      }
      else
      {
         predictions_update_cols = 3;
         prediction_fn_value_col = 2;
      }
      if( second_variable_to_predict_over != variable_to_predict_over )
      {
         predictions_update_cols++;
         prediction_fn_value_col++;
      }
      if( predict_on_data )
         predictions_mx_rows = n;
      else if( n > regression_n )
         predictions_mx_rows = n - regression_n;
      else
         predictions_mx_rows = total_num_knot_point_candidates;
      if( extra_deleted_point )
         predictions_mx_rows++;
      predictions_mx_cols = predictions_update_cols + 1;
      predictions_mx = new double*[ predictions_mx_rows ];
      for( i = 0; i < predictions_mx_rows; i++ )
         predictions_mx[ i ] = new double[ predictions_mx_cols ];
      for( i = 0; i < predictions_mx_rows; i++ )
      {
         if( extra_deleted_point && i == delete_point )
         {
            x_val = deleted_x_point[ variable_to_predict_over ];
//            shift--;
         }
         else if( !predict_on_knot_point_candidates && test_grid_points )
            x_val = test_grid_points[ i ][ variable_to_predict_over ];
//            x_val = gsl_matrix_get( X, shift + i, variable_to_predict_over );
         else if( predict_on_knot_point_candidates )
            x_val = knot_point_candidates[ variable_to_predict_over ][ i ];
         if( normalise_predictors && x_scale[ variable_to_predict_over ] )
         {
            x_val *= x_scale[ variable_to_predict_over ];
            if( normalise_predictors == 1 )
               x_val += x_min[ variable_to_predict_over ];
         }
  	 if( logistic_transform_predictors )
	    x_val = logit_transform( x_val );
         predictions_mx[ i ][ 0 ] = x_val;
         if( variable_to_predict_over != second_variable_to_predict_over )
         {
            if( extra_deleted_point && i == delete_point )
               x_val = deleted_x_point[ second_variable_to_predict_over ];
            else if( !predict_on_knot_point_candidates )
               x_val = test_grid_points[ i ][ second_variable_to_predict_over ];
//               x_val = gsl_matrix_get( X, shift + i, second_variable_to_predict_over );
            else
               x_val = knot_point_candidates[ second_variable_to_predict_over ][ i ];
            if( normalise_predictors && x_scale[ second_variable_to_predict_over ] )
            {
               x_val *= x_scale[ second_variable_to_predict_over ];
               if( normalise_predictors == 1 )
                  x_val += x_min[ second_variable_to_predict_over ];
            }
            predictions_mx[ i ][ 1 ] = x_val;
         }
      }
//      if( extra_deleted_point )
//         shift++;
   }

//   if( !storing_basis_function_coefficients )//
   //put the posterior means in for basis coefficients

   if( predictions_mx )
   {
      last_col = predictions_mx_cols - 1;
      double fn_val = 0, int_fn_val, sd_val = 0, *xx = 0;//new double[ p ];
      for( i = 0; i < predictions_mx_rows; i++ )
      {
         if( extra_deleted_point && i == delete_point )
         {
            for( j = 0; j < p; j++ )
               xx = deleted_x_point;
//               xx[ j ] = deleted_x_point[ j ];
//            shift--;
         }
         else
//            for( j = 0; j < p; j++ )
               if( !predict_on_knot_point_candidates )
                  xx = test_grid_points[ i ];
//                  xx[ j ] = gsl_matrix_get( X, shift + i, j );
               else
                  xx = knot_point_candidates[ i ];
//                  xx[ j ] = knot_point_candidates[ i ][ j ];
         fn_val = function_evaluate( xx );
         int_fn_val = integral_function_evaluate( xx, 1, variable_to_predict_over );
         if( predictions_update_cols == 3 )
            sd_val = 0;//sqrt( function_variance_evaluate( xx ) );
         if( normalise_targets && y_scale[ 0 ] > 0 )
         {
            fn_val = ( fn_val + y_bar[ 0 ] ) * y_scale[ 0 ];
            sd_val *= y_scale[ 0 ];
         }
	 //         if( log_targets )
	 //            fn_val = exp( fn_val ) - 1;
	 double x_val = test_grid_points[ i ][ variable_to_predict_over ];
         if( log_inputs )
            fn_val /= predictions_mx[ i ][ 0 ];
	 else if( logistic_transform_predictors )
	    fn_val *= x_val*(1-x_val);
         if( predictions_update_cols == 3 )
         {
             predictions_mx[ i ][ last_col - 2 ] = 0;//fn_val - 2 * sd_val;
            predictions_mx[ i ][ last_col - 1 ] = fn_val;
            predictions_mx[ i ][ last_col ] = int_fn_val;//fn_val + 2 * sd_val;
         }
         else
            predictions_mx[ i ][ last_col ] = fn_val;
      }
//      delete [] xx;
   }
   //   if( !storing_basis_function_coefficients )
   //      revert_basis_coefficients();
}

double MCMC_Density_Estimation::prior_measure( double left_t, double right_t )
{
  //   return right_t - left_t;
   if( !normalise_predictors )
   {
      left_t = ( left_t - x_min[ 0 ] ) / x_scale[ 0 ];
      right_t = ( right_t - x_min[ 0 ] ) / x_scale[ 0 ];
   }
   double measure = gsl_cdf_beta_P( right_t, beta_distn_alpha, beta_distn_beta ) - gsl_cdf_beta_P( left_t, beta_distn_alpha, beta_distn_beta );
   if( !normalise_predictors )
      measure *= x_scale[ 0 ];
   return measure;
}

double MCMC_Density_Estimation::calculate_log_likelihood()
{
   proposed_log_likelihood = ml_const;
   proposed_partition_point_count.clear();
   proposed_partition_volumes.clear();
   proposed_partition_measures.clear();
   proposed_partition_starting_points.clear();
   proposed_partition_posterior_means.clear();
   proposed_partition_mls.clear();
   proposed_partition_posterior_means.clear();
   int i, j;
   int x_start = 0, x_end = regression_n;
   for( j = 0; j < p; j++ )
   {
       Basis_Fn* left_basis = get_first_unremovable_basis_fn( j );
       double left_t = left_basis->get_basis_fn_location()[ 0 ];
       double maximum_t = normalise_predictors == 1 ? 1 : x_max[ j ];
       double right_t = maximum_t;
       Basis_Fn* right_basis = get_first_basis_fn_to_the_right( left_basis );
       double width = 0, measure = 0;
       list<int> points_inside;
       int num_points_inside = 0;
       bool reached_end = false;
       while( !reached_end )
       {
           reached_end = right_basis == left_basis;
           if( !reached_end )
              right_t = right_basis->get_basis_fn_location()[ 0 ];
           else
              right_t = maximum_t;
	   width = right_t - left_t;
           measure = prior_measure( left_t, right_t );
           i = x_start;
	   bool not_finished = true;
           while( i < x_end && not_finished )
           {
              double x_ij = gsl_matrix_get( X, i, j );
              if( x_ij >= left_t && (reached_end || x_ij < right_t) )
              {
                 points_inside.push_back( i );
                 num_points_inside++;
              }
	      else if( x_sorted )
	      {
		 not_finished = false;
		 x_start = i;
	      }
              i++;
           }
	   proposed_partition_volumes.push_back( width );
	   proposed_partition_measures.push_back( measure );
	   proposed_partition_point_count.push_back( num_points_inside );
           proposed_partition_starting_points.push_back( points_inside.front() );
	   double post_mean = (alpha * measure + num_points_inside ) / dirichlet_normaliser;
	   proposed_partition_posterior_means.push_back( post_mean );
           double ml = 0;
	   if( num_points_inside == 1 )
	      ml = log_alpha;
	   else if( num_points_inside > 1 )
	   {
	      ml = -num_points_inside * log( width );
	      for( i = 0; i < num_points_inside; i++ )
		 ml += log( alpha * measure + i );
	   }
           proposed_log_likelihood += ml;
           proposed_partition_mls.push_back( ml );
           points_inside.clear();
           num_points_inside = 0;
           left_basis = right_basis;
           left_t = right_t;
           right_basis = get_first_basis_fn_to_the_right( left_basis );
       }
   }
   number_of_models_evaluated++;
   return proposed_log_likelihood;
}

void MCMC_Density_Estimation::set_basis_coeffs()
{
   list<int>::iterator count_iter = partition_point_count.begin();
   list<double>::iterator vol_iter = partition_volumes.begin();
   list<double>::iterator mean_iter = partition_posterior_means.begin();
   double left_mean = 0, right_mean = 0;
   for( int j = 0; j < p; j++ )
   {
       bool reached_end = false;
       Basis_Fn* left_basis = get_first_unremovable_basis_fn( j );
       Basis_Fn* right_basis = get_first_basis_fn_to_the_right( left_basis );
       while( !reached_end )
       {
	  reached_end = right_basis == left_basis;
	  //	  right_mean = (alpha * *vol_iter + *count_iter ) / (dirichlet_normaliser * *vol_iter );
	  right_mean = *mean_iter / *vol_iter;
          double coeff = right_mean - left_mean;
          if( normalise_predictors )
           coeff /= x_scale[ j ];
          left_basis->set_coeff( coeff );
	  if( !reached_end )
	  {
	     left_mean = right_mean;
	     left_basis = right_basis;
	     right_basis = get_first_basis_fn_to_the_right( left_basis );
	     ++count_iter;
	     ++vol_iter;
	     ++mean_iter;
	  }
       }
   }
}

void MCMC_Density_Estimation::update_posterior_maximizing_model()
{
   best_changepoints.clear();
   best_components.clear();
   best_changepoint_fn_vals.clear();
   best_changepoint_unremovable.clear();
   for( int j = 0; j < p; j++ )
   {
       Basis_Fn* left_basis = NULL;
       Basis_Fn* right_basis = get_first_unremovable_basis_fn( j );
       while( right_basis != left_basis )
       {
	  double basis_location = right_basis->get_basis_fn_location()[ 0 ];
	  int basis_component = right_basis->get_basis_fn_component()[ 0 ];
	  double basis_coeff = right_basis->get_coeff();
	  best_changepoints.push_back( basis_location );
	  best_components.push_back( basis_component );
	  best_changepoint_fn_vals.push_back( basis_coeff );
          best_changepoint_unremovable.push_back( right_basis == get_first_unremovable_basis_fn( basis_component ) );
	  left_basis = right_basis;
	  right_basis = get_first_basis_fn_to_the_right( left_basis );
       }
   }
   write_posterior_maximizing_model();
   if( !model_averaging )
      make_predictions();
}

void MCMC_Density_Estimation::write_posterior_maximizing_model()
{
   if( !strlen( Best_Changepoints_Filename ) )
      return;
   ofstream BestChangePointsStream;
   BestChangePointsStream.open( Best_Changepoints_Filename, ios::out );
   if( !BestChangePointsStream.is_open() )
      return;
   list<double>::iterator cp_iter = best_changepoints.begin();
   list<int>::iterator comp_iter = best_components.begin();
   list<double>::iterator fn_iter = best_changepoint_fn_vals.begin();
   list<bool>::iterator unremov_iter = best_changepoint_unremovable.begin();
   while( cp_iter != best_changepoints.end() )
   {
      double cp = *cp_iter;
      if( normalise_predictors ) cp *= x_scale[ *comp_iter ];
      if( normalise_predictors == 1 ) cp += x_min[ *comp_iter ];
      if( logistic_transform_predictors )
      {
	 if( cp <= 0 )
	    cp += 0.001;
	 cp = logit_transform( cp );
      }
      BestChangePointsStream << *comp_iter << "\t" << cp << "\t" << *fn_iter << "\t" << *unremov_iter << endl;
      ++cp_iter;
      ++comp_iter;
      ++fn_iter;
      ++unremov_iter;
   }
}


void MCMC_Density_Estimation::revise_log_posterior()
{
   list<int>::iterator count_iter = proposed_partition_point_count.begin();
   list<double>::iterator vol_iter = proposed_partition_volumes.begin();
   list<double>::iterator measure_iter = proposed_partition_measures.begin();
   list<int>::iterator x_iter = proposed_partition_starting_points.begin();
   list<double>::iterator mean_iter = proposed_partition_posterior_means.begin();
   list<double>::iterator ml_iter = proposed_partition_mls.begin();
   if( p != 1 ){ calculate_log_posterior(); switch_values( log_posterior, proposed_log_posterior ); return; }

   double left_width = 0, left_measure = 0, left_mean = 0;
   int i = 0, left_num_points_inside = 0;
   double left_ml = 0, right_ml = 0;
   //   if( move_type == ADD_BASIS || move_type == DEL_BASIS || move_type == MOVE_BASIS )
   //   {
      int comp = (move_type == DEL_BASIS ? Removed_Basis : New_Basis )->get_basis_fn_component()[ 0 ];
   Basis_Fn* bas = get_first_unremovable_basis_fn( comp );
   while( bas != Basis_to_the_left )
//   int pos = position_in_basis_number_to_pointer( Basis_to_the_left );
//   for( int ii = 0; ii < pos; ii++ )
   {
       bas = get_first_basis_fn_to_the_right( bas );
       ++count_iter; ++vol_iter; ++measure_iter; ++x_iter; ++mean_iter; ++ml_iter;
   }
   if( move_type == ADD_BASIS || move_type == MOVE_BASIS )
   {
      double left_t = Basis_to_the_left->get_basis_fn_location()[ 0 ];
      double right_t = New_Basis->get_basis_fn_location()[ 0 ];
      left_measure = prior_measure( left_t, right_t );
      left_width = right_t - left_t;
      list<int> points_inside;
      int x_start = 0;
      i = x_sorted ? *x_iter : 0;
      bool not_finished = true;
      while( i < regression_n && not_finished )
      {
         double x_ij = gsl_matrix_get( X, i, comp );
         if( x_ij >= left_t && x_ij < right_t )
         {
            points_inside.push_back( i );
            left_num_points_inside++;
         }
         else if( x_sorted && x_ij >= right_t )
         {
            not_finished = false;
            x_start = i;
         }
         i++;
      }
      ++x_iter;
      proposed_partition_starting_points.insert( x_iter, x_start );
   }
   if( move_type == DEL_BASIS )
      ++x_iter;
      if( move_type == DEL_BASIS || move_type == MOVE_BASIS )
         proposed_partition_starting_points.erase( x_iter );
      int old_num_points_inside = *count_iter;
      if( move_type == DEL_BASIS || move_type == MOVE_BASIS )
      {
          count_iter = proposed_partition_point_count.erase( count_iter );
          old_num_points_inside += *count_iter;
      }
      int right_num_points_inside = old_num_points_inside - left_num_points_inside;
///      double b_right = Basis_to_the_right ? (Basis_to_the_right->get_basis_fn_location()[ 0 ]):100;
      if( move_type != DEL_BASIS )
	 proposed_partition_point_count.insert( count_iter, left_num_points_inside );
      proposed_partition_point_count.insert( count_iter, right_num_points_inside );
      proposed_partition_point_count.erase( count_iter );
//      *count_iter = left_num_points_inside;
      double old_volume = *vol_iter;
      double old_measure = *measure_iter;
      double old_mean = *mean_iter;
      if( move_type == DEL_BASIS || move_type == MOVE_BASIS )
      {
          vol_iter = proposed_partition_volumes.erase( vol_iter );
          old_volume += *vol_iter;
          measure_iter = proposed_partition_volumes.erase( measure_iter );
          old_measure += *measure_iter;
	  mean_iter = proposed_partition_posterior_means.erase( mean_iter );
	  old_mean += *mean_iter;
      }
      double right_width = old_volume - left_width;
      double right_measure = old_measure - left_measure;
      left_mean = (alpha * left_measure + left_num_points_inside ) / dirichlet_normaliser;
      double right_mean = old_mean - left_mean;
      if( move_type != DEL_BASIS )
      {
	 proposed_partition_volumes.insert( vol_iter, left_width );
	 proposed_partition_measures.insert( measure_iter, left_measure );
	 proposed_partition_posterior_means.insert( mean_iter, left_mean );
      }
      proposed_partition_volumes.insert( vol_iter, right_width );
      proposed_partition_volumes.erase( vol_iter );
      proposed_partition_measures.insert( measure_iter, right_measure );
      proposed_partition_measures.erase( measure_iter );
      proposed_partition_posterior_means.insert( mean_iter, right_mean );
      proposed_partition_posterior_means.erase( mean_iter );
//      *vol_iter = left_width;
      double old_ml = *ml_iter;
      if( move_type == DEL_BASIS || move_type == MOVE_BASIS )
      {
          ml_iter = proposed_partition_mls.erase( ml_iter );
          old_ml += *ml_iter;
      }
      if( move_type != DEL_BASIS )
      {
	  if( left_num_points_inside == 1 )
	     left_ml = log_alpha;
	  else if( left_num_points_inside > 1 )
	  {
	     left_ml = -left_num_points_inside * log( left_width );
	     for( i = 0; i < left_num_points_inside; i++ )
	        left_ml += log( alpha * left_measure + i );
	  }
	 proposed_partition_mls.insert( ml_iter, left_ml );
      }
      if( right_num_points_inside == 1 )
	 right_ml = log_alpha;
      else if( right_num_points_inside > 1 )
      {
	 right_ml = -right_num_points_inside * log( right_width );
	 for( i = 0; i < right_num_points_inside; i++ )
	    right_ml += log( alpha * right_measure + i );
      }
      proposed_partition_mls.insert( ml_iter, right_ml );
      proposed_partition_mls.erase( ml_iter );
      proposed_log_likelihood += left_ml + right_ml - old_ml;
      if( move_type == ADD_BASIS || move_type == DEL_BASIS )
         calculate_log_prior();
      proposed_log_posterior = proposed_log_prior + proposed_log_likelihood;
      number_of_models_evaluated++;
      //   }
   //   else
   //   {
   //      calculate_log_posterior();
   //      switch_values( log_posterior, proposed_log_posterior );
   //   }
      
}

void MCMC_Density_Estimation::end_of_chain_processing()
{
   write_knot_point_frequencies();
}
